import torch

from torchrua.ty import C, D, P, T


def last_c(sequence: C) -> T:
    data, token_sizes = sequence
    return data[token_sizes.cumsum(dim=0) - 1]


C.last = last_c


def last_d(sequence: D) -> T:
    b, t, *_ = sequence.size()
    batch_ptr = torch.arange(b, dtype=torch.long, device=sequence.data.device)
    return sequence.data[batch_ptr, sequence.token_sizes - 1]


D.last = last_d


def last_p(sequence: P) -> T:
    return sequence.idx().pad().last().rua(sequence)


P.last = last_p



from torchrua.catting import *
from torchrua.compose import *
from torchrua.core import *
from torchrua.decode import *
from torchrua.head import *
from torchrua.last import *
from torchrua.mask import *
from torchrua.packing import *
from torchrua.padding import *
from torchrua.primitives import *
from torchrua.reduce import *
from torchrua.reverse import *
from torchrua.roll import *
from torchrua.segment import *
from torchrua.trunc import *
from torchrua.ty import *



from typing import List

import torch
from torch.nn.utils.rnn import PackedSequence

from torchrua.core import _self, invert_permutation
from torchrua.ty import C, D, P, T


def pack_sequence(sequence: List[T]) -> P:
    return C.new(sequence).pack()


P.new = pack_sequence
P.pack = _self


def pack_t(sequence: T) -> P:
    batch_sizes = torch.ones(sequence.size()[:1], dtype=torch.long).cpu()
    sorted_indices = sequence.data.new_tensor([0], dtype=torch.long)
    unsorted_indices = sequence.data.new_tensor([0], dtype=torch.long)

    return PackedSequence(
        data=sequence.data,
        batch_sizes=batch_sizes,
        sorted_indices=sorted_indices,
        unsorted_indices=unsorted_indices,
    )


T.pack = pack_t


def pack_c(sequence: C) -> P:
    data, token_sizes = sequence
    b, t, *sizes = sequence.size()

    if len(sizes) > 0:
        return sequence.idx().pack().rua(sequence)

    _, sorting_indices = torch.sort(token_sizes.detach().cpu(), descending=True)
    sorting_indices = sorting_indices.to(device=data.device)
    unsorted_indices = invert_permutation(sorting_indices)

    batch_ptr, token_ptr = sequence.ptr()
    batch_ptr = unsorted_indices[batch_ptr]

    tensor = data.new_zeros((t, b))
    tensor[token_ptr, batch_ptr] = data

    mask = torch.zeros_like(tensor, dtype=torch.long)
    mask[token_ptr, batch_ptr] = 1

    return PackedSequence(
        data=tensor[mask.bool()],
        batch_sizes=mask.sum(dim=1).detach().cpu(),
        sorted_indices=sorting_indices,
        unsorted_indices=unsorted_indices,
    )


C.pack = pack_c


def pack_d(sequence: D) -> P:
    return sequence.idx().pack().rua(sequence)


D.pack = pack_d



from typing import Union

import torch
from torch.types import Number

from torchrua.ty import C, D, P, T


def mask_cd(sequence: Union[C, D, P], zero: Number = False, one: Number = True, dtype: torch.dtype = torch.bool) -> T:
    b, t, *_ = sequence.size()
    batch_ptr, token_ptr = sequence.ptr()

    mask = sequence.data.new_full((b, t), fill_value=zero, dtype=dtype)
    mask[batch_ptr, token_ptr] = one
    return mask


C.mask = mask_cd
D.mask = mask_cd


def mask_p(sequence: Union[C, D, P], zero: Number = False, one: Number = True, dtype: torch.dtype = torch.bool) -> T:
    b, t, *_ = sequence.size()
    batch_ptr, token_ptr = sequence.ptr()

    mask = sequence.data.new_full((b, t), fill_value=zero, dtype=dtype)
    mask[sequence.sorted_indices[batch_ptr], token_ptr] = one
    return mask


P.mask = mask_p



from typing import Any, NamedTuple, Type, Union

import torch
from torch import Tensor
from torch.nn.utils.rnn import PackedSequence


class CattedSequence(NamedTuple):
    data: Tensor
    token_sizes: Tensor

    def to(self, dtype: torch.dtype = None, device: torch.device = None) -> 'CattedSequence':
        return CattedSequence(
            data=self.data.to(dtype=dtype, device=device),
            token_sizes=self.token_sizes.to(dtype=torch.long, device=device),
        )

    def double(self) -> 'CattedSequence':
        return self.to(dtype=torch.double)

    def float(self) -> 'CattedSequence':
        return self.to(dtype=torch.float)

    def half(self) -> 'CattedSequence':
        return self.to(dtype=torch.half)

    def long(self) -> 'CattedSequence':
        return self.to(dtype=torch.long)

    def int(self) -> 'CattedSequence':
        return self.to(dtype=torch.int)

    def short(self) -> 'CattedSequence':
        return self.to(dtype=torch.short)

    def char(self) -> 'CattedSequence':
        return self.to(dtype=torch.int8)

    def byte(self) -> 'CattedSequence':
        return self.to(dtype=torch.uint8)

    def cpu(self) -> 'CattedSequence':
        return CattedSequence(
            data=self.data.cpu(),
            token_sizes=self.token_sizes.cpu(),
        )

    def cuda(self) -> 'CattedSequence':
        return CattedSequence(
            data=self.data.cuda(),
            token_sizes=self.token_sizes.cuda(),
        )

    def detach(self) -> 'CattedSequence':
        return CattedSequence(
            data=self.data.detach(),
            token_sizes=self.token_sizes.detach(),
        )


class PaddedSequence(NamedTuple):
    data: Tensor
    token_sizes: Tensor

    def to(self, dtype: torch.dtype = None, device: torch.device = None) -> 'PaddedSequence':
        return PaddedSequence(
            data=self.data.to(dtype=dtype, device=device),
            token_sizes=self.token_sizes.to(dtype=torch.long, device=device),
        )

    def double(self) -> 'PaddedSequence':
        return self.to(dtype=torch.double)

    def float(self) -> 'PaddedSequence':
        return self.to(dtype=torch.float)

    def half(self) -> 'PaddedSequence':
        return self.to(dtype=torch.half)

    def long(self) -> 'PaddedSequence':
        return self.to(dtype=torch.long)

    def int(self) -> 'PaddedSequence':
        return self.to(dtype=torch.int)

    def short(self) -> 'PaddedSequence':
        return self.to(dtype=torch.short)

    def char(self) -> 'PaddedSequence':
        return self.to(dtype=torch.int8)

    def byte(self) -> 'PaddedSequence':
        return self.to(dtype=torch.uint8)

    def cpu(self) -> 'PaddedSequence':
        return PaddedSequence(
            data=self.data.cpu(),
            token_sizes=self.token_sizes.cpu(),
        )

    def cuda(self) -> 'PaddedSequence':
        return PaddedSequence(
            data=self.data.cuda(),
            token_sizes=self.token_sizes.cuda(),
        )

    def detach(self) -> 'PaddedSequence':
        return PaddedSequence(
            data=self.data.detach(),
            token_sizes=self.token_sizes.detach(),
        )


T = Tensor
C = CattedSequence
D = PaddedSequence
P = PackedSequence


def is_type(obj: Any, ty: Type) -> bool:
    __origin__ = getattr(ty, '__origin__', None)
    __args__ = getattr(ty, '__args__', [])

    if __origin__ is list:
        if not isinstance(obj, list):
            return False

        return all(is_type(o, __args__[0]) for o in obj)

    if __origin__ is tuple:
        if isinstance(obj, (C, D, P)) or not isinstance(obj, tuple):
            return False

        if len(__args__) == 2 and __args__[1] is ...:
            return all(is_type(o, __args__[0]) for o in obj)

        if len(__args__) == len(obj):
            return all(is_type(o, t) for o, t in zip(obj, __args__))

        return False

    if __origin__ is Union:
        return any(is_type(obj, t) for t in __args__)

    return isinstance(obj, ty)


def idx_cp(sequence: Union[C, P]) -> Union[C, P]:
    n, *_ = sequence.data.size()
    data = torch.arange(n, dtype=torch.long, device=sequence.data.device)
    return sequence._replace(data=data)


def idx_d(sequence: D) -> C:
    _, token_sizes = sequence
    _, t, *_ = sequence.size()
    batch_ptr, token_ptr = sequence.ptr()
    return CattedSequence(data=token_ptr + batch_ptr * t, token_sizes=token_sizes)


C.idx = idx_cp
D.idx = idx_d
P.idx = idx_cp


def _data_t(sequence: T) -> T:
    return sequence


def _data_d(sequence: D) -> T:
    return sequence.data.flatten(start_dim=0, end_dim=1)


def _data_cp(sequence: Union[C, P]) -> T:
    return sequence.data


T._data = _data_t
C._data = _data_cp
D._data = _data_d
P._data = _data_cp


def _replace_t(_: T, data: T) -> T:
    return data


T._replace = _replace_t


def rua(index: Union[T, C, P], sequence: Union[T, C, D, P], *indices: Union[T, C, P]) -> Union[T, C, P]:
    indices = tuple(index.data for index in (index, *indices))
    return index._replace(data=sequence._data()[indices])


T.rua = rua
C.rua = rua
P.rua = rua


def empty_cdp(sequence: Union[C, D, P]) -> Union[C, D, P]:
    return sequence._replace(data=sequence.data.new_tensor(()))


C.empty = empty_cdp
D.empty = empty_cdp
P.empty = empty_cdp



from typing import List

import torch
from torch.types import Number

from torchrua.core import _self
from torchrua.ty import C, D, P, T


def pad_sequence(sequence: List[T], fill_value: Number = 0) -> D:
    return C.new(sequence).pad(fill_value=fill_value)


D.new = pad_sequence
D.pad = _self


def pad_d(sequence: T, fill_value: Number = 0) -> D:
    token_sizes = sequence.new_tensor(sequence.size()[:1], dtype=torch.long)
    return D(data=sequence[None], token_sizes=token_sizes)


T.pad = pad_d


def pad_c(sequence: C, fill_value: Number = 0) -> D:
    data, token_sizes = sequence

    b, t, *sizes = sequence.size()
    batch_ptr, token_ptr = sequence.ptr()

    tensor = data.new_full((b, t, *sizes), fill_value=fill_value)
    tensor[batch_ptr, token_ptr] = data

    return D(data=tensor, token_sizes=token_sizes)


C.pad = pad_c


def pad_p(sequence: P, fill_value: Number = 0) -> D:
    data, _, sorted_indices, _ = sequence

    b, t, *sizes = sequence.size()
    batch_ptr, token_ptr = sequence.ptr()
    batch_ptr = sorted_indices[batch_ptr]

    tensor = data.new_full((b, t, *sizes), fill_value=fill_value)
    tensor[batch_ptr, token_ptr] = data

    mask = data.new_zeros((b, t), dtype=torch.long)
    mask[batch_ptr, token_ptr] = 1

    return D(data=tensor, token_sizes=mask.sum(dim=1))


P.pad = pad_p



from typing import Tuple

import torch

from torchrua.core import major_sizes_to_ptr
from torchrua.ty import C, D, P


def trunc_c(sequence: C, trunc: Tuple[int, int]) -> C:
    data, token_sizes = sequence

    token_sizes = torch.stack([
        torch.full_like(token_sizes, fill_value=trunc[0]),
        token_sizes - trunc[0] - trunc[1],
        torch.full_like(token_sizes, fill_value=trunc[1]),
    ], dim=-1)

    data = torch.split(data, token_sizes.view(-1).cpu().tolist(), dim=0)

    return C(data=torch.cat(data[1::3]), token_sizes=token_sizes[:, 1])


C.trunc = trunc_c


def trunc_d(sequence: D, trunc: Tuple[int, int]) -> D:
    data, token_sizes = sequence
    _, t, *_ = sequence.size()

    return D(
        data=data[:, trunc[0]:t - trunc[1]],
        token_sizes=token_sizes - trunc[0] - trunc[1],
    )


D.trunc = trunc_d


def trunc_p(sequence: P, trunc: Tuple[int, int]) -> P:
    data, batch_sizes, _, _ = sequence

    batch_sizes = batch_sizes[trunc[0] + trunc[1]:]
    batch_ptr, token_ptr = major_sizes_to_ptr(sizes=batch_sizes.to(device=data.device))
    index = batch_ptr + sequence.offsets()[token_ptr + trunc[0]]

    return sequence._replace(data=data[index], batch_sizes=batch_sizes)


P.trunc = trunc_p



import torch
from hypothesis import given, settings, strategies as st
from torch import nn
from torch.nn.utils.rnn import pack_sequence
from torchnyan import FEATURE_DIM, TINY_BATCH_SIZE, TINY_TOKEN_SIZE, assert_close, assert_grad_close, device, sizes

from torchrua import C, D, P, compose


@settings(deadline=None)
@given(
    data=st.data(),
    token_sizes_batch=sizes(TINY_BATCH_SIZE, TINY_BATCH_SIZE, TINY_TOKEN_SIZE),
    input_size=sizes(FEATURE_DIM),
    hidden_size=sizes(FEATURE_DIM),
)
def test_compose_sequences(data, token_sizes_batch, input_size, hidden_size):
    sequences = [
        [
            torch.randn((token_size, input_size), requires_grad=True, device=device)
            for token_size in token_sizes
        ]
        for token_sizes in token_sizes_batch
    ]

    rnn = nn.LSTM(
        input_size=input_size,
        hidden_size=hidden_size,
        bidirectional=True, bias=True,
    ).to(device=device)

    actual_sequences = [
        data.draw(st.sampled_from([C, D, P])).new(sequence).to(device=device)
        for sequence in sequences
    ]
    _, (actual, _) = rnn(compose(actual_sequences))
    actual = actual.transpose(-3, -2).flatten(start_dim=-2)

    expected = []
    for sequence in sequences:
        _, (hidden, _) = rnn(pack_sequence(sequence, enforce_sorted=False))
        expected.append(hidden.transpose(-3, -2).flatten(start_dim=-2))
    expected, _, _, _ = pack_sequence(expected, enforce_sorted=False)

    assert_close(actual, expected, check_stride=False)
    assert_grad_close(actual, expected, inputs=[token for sequence in sequences for token in sequence])



from typing import Tuple, Union

import torch

from torchrua.core import get_offsets, major_sizes_to_ptr
from torchrua.ty import C, D, P, T, is_type


def size_c(sequence: C) -> Tuple[int, ...]:
    data, token_sizes = sequence
    return token_sizes.size()[0], token_sizes.max().item(), *data.size()[1:]


C.size = size_c


def size_d(sequence: D) -> Tuple[int, ...]:
    data, token_sizes = sequence
    return token_sizes.size()[0], token_sizes.max().item(), *data.size()[2:]


D.size = size_d


def size_p(sequence: P) -> Tuple[int, ...]:
    data, batch_sizes, _, _ = sequence
    return batch_sizes.max().item(), batch_sizes.size()[0], *data.size()[1:]


P.size = size_p


def ptr_cd(sequence: Union[C, D]) -> Tuple[T, T]:
    _, token_sizes = sequence
    token_ptr, batch_ptr = major_sizes_to_ptr(sizes=token_sizes)
    return batch_ptr, token_ptr


C.ptr = ptr_cd
D.ptr = ptr_cd


def ptr_p(sequence: P) -> Tuple[T, T]:
    data, batch_sizes, _, _ = sequence
    batch_ptr, token_ptr = major_sizes_to_ptr(sizes=batch_sizes.to(device=data.device))
    return batch_ptr, token_ptr


P.ptr = ptr_p


def offsets_c(sequence: C) -> T:
    _, token_sizes = sequence
    return get_offsets(token_sizes)


C.offsets = offsets_c


def offsets_d(sequence: D) -> T:
    data, token_sizes = sequence
    b, t = sequence.size()
    return torch.arange(b, dtype=torch.long, device=data.device) * t


D.offsets = offsets_d


def offsets_p(sequence: P) -> T:
    data, batch_sizes, _, _ = sequence
    return get_offsets(batch_sizes.to(device=data.device))


P.offsets = offsets_p


def __getitem__c(sequence: C, ptr: Tuple[T, T]) -> C:
    if is_type(ptr, Tuple[T, T]):
        batch_ptr, token_ptr = ptr
        return sequence._replace(data=sequence.data[sequence.offsets()[batch_ptr] + token_ptr])

    return super(C, sequence).__getitem__(ptr)


C.__getitem__ = __getitem__c


def __getitem__d(sequence: D, ptr: Tuple[T, T]) -> D:
    if is_type(ptr, Tuple[T, T]):
        batch_ptr, token_ptr = ptr
        return sequence._replace(data=sequence.data[batch_ptr, token_ptr])

    return super(D, sequence).__getitem__(ptr)


D.__getitem__ = __getitem__d


def __getitem__p(sequence: P, ptr: Tuple[T, T]) -> P:
    if is_type(ptr, Tuple[T, T]):
        batch_ptr, token_ptr = ptr
        return sequence._replace(data=sequence.data[batch_ptr + sequence.offsets()[token_ptr]])

    return super(P, sequence).__getitem__(ptr)


P.__getitem__ = __getitem__p



import torch

from torchrua.core import get_offsets
from torchrua.ty import C, D, P, T


def head_c(sequence: C) -> T:
    data, token_sizes = sequence
    m, *_ = data.size()

    idx = get_offsets(sizes=token_sizes)
    return data[torch.clamp_max_(idx, max=m - 1)]


C.head = head_c


def head_d(sequence: D) -> T:
    data, _ = sequence
    return data[:, 0]


D.head = head_d


def head_p(sequence: P) -> T:
    data, batch_sizes, _, unsorted_indices = sequence
    if unsorted_indices is not None:
        return data[unsorted_indices]
    return data[:batch_sizes[0].detach().cpu().item()]


P.head = head_p



import torch

from torchrua.ty import C, T


def scatter_max(tensor: T, index: T, source: T, include_self: bool = False, dim: int = 0):
    return torch.index_reduce(tensor, index=index, source=source, reduce='amax', include_self=include_self, dim=dim)


def scatter_min(tensor: T, index: T, source: T, include_self: bool = False, dim: int = 0):
    return torch.index_reduce(tensor, index=index, source=source, reduce='amin', include_self=include_self, dim=dim)


def scatter_sum(tensor: T, index: T, source: T, include_self: bool = False, dim: int = 0):
    return torch.index_add(tensor if include_self else torch.zeros_like(tensor), index=index, source=source, dim=dim)


def scatter_mean(tensor: T, index: T, source: T, include_self: bool = False, dim: int = 0):
    return torch.index_reduce(tensor, index=index, source=source, reduce='mean', include_self=include_self, dim=dim)


def scatter_prod(tensor: T, index: T, source: T, include_self: bool = False, dim: int = 0):
    return torch.index_reduce(tensor, index=index, source=source, reduce='prod', include_self=include_self, dim=dim)


def scatter_logsumexp(tensor: T, index: T, source: T, include_self: bool = False, dim: int = 0):
    m = scatter_max(tensor, index=index, source=source, include_self=include_self, dim=dim).detach()

    tensor = (tensor - m).exp()
    source = (source - m[index]).exp()
    return scatter_sum(tensor, index=index, source=source, include_self=include_self, dim=dim).log() + m


def segment_max(tensor: T, segment_sizes: T) -> T:
    m = tensor.min().detach().cpu().item()
    return torch.segment_reduce(tensor, reduce='max', lengths=segment_sizes, unsafe=True, initial=m)


def segment_min(tensor: T, segment_sizes: T) -> T:
    m = tensor.max().detach().cpu().item()
    return torch.segment_reduce(tensor, reduce='min', lengths=segment_sizes, unsafe=True, initial=m)


def segment_sum(tensor: T, segment_sizes: T) -> T:
    return torch.segment_reduce(tensor, reduce='sum', lengths=segment_sizes, unsafe=True, initial=0)


def segment_mean(tensor: T, segment_sizes: T) -> T:
    return torch.segment_reduce(tensor, reduce='mean', lengths=segment_sizes, unsafe=True, initial=0)


def segment_prod(tensor: T, segment_sizes: T) -> T:
    return torch.segment_reduce(tensor, reduce='prod', lengths=segment_sizes, unsafe=True, initial=1)


def segment_logsumexp(tensor: T, segment_sizes: T) -> T:
    m = segment_max(tensor, segment_sizes=segment_sizes).detach()

    tensor = (tensor - torch.repeat_interleave(m, dim=0, repeats=segment_sizes)).exp()
    eps = (segment_sizes == 0).to(dtype=tensor.dtype).view((-1, *[1 for _ in tensor.size()[1:]]))
    return (segment_sum(tensor, segment_sizes=segment_sizes) + eps).log() + m


def segment_head(tensor: T, segment_sizes: T) -> T:
    return C(data=tensor, token_sizes=segment_sizes).head()


def segment_last(tensor: T, segment_sizes: T) -> T:
    return C(data=tensor, token_sizes=segment_sizes).last()



import torch

from torchrua.ty import C, D, P


def roll_c(sequence: C, shifts: int) -> C:
    data, token_sizes = sequence

    batch_ptr, token_ptr = sequence.ptr()
    token_sizes = torch.repeat_interleave(token_sizes, token_sizes)
    token_ptr = (token_ptr - shifts + token_sizes) % token_sizes

    return sequence[batch_ptr, token_ptr]


C.roll = roll_c


def roll_d(sequence: D, shifts: int) -> D:
    index1, _ = idx = sequence.idx()
    index2, _ = idx.roll(shifts)

    data = sequence._data().clone()
    data[index1] = data[index2]

    return sequence._replace(data=data.view_as(sequence.data))


D.roll = roll_d


def roll_p(sequence: P, shifts: int) -> P:
    return sequence.idx().cat().roll(shifts).pack().rua(sequence)


P.roll = roll_p



import torch
from hypothesis import given, settings, strategies as st
from torch import Tensor
from torchnyan import BATCH_SIZE, FEATURE_DIM, TOKEN_SIZE, assert_grad_close, assert_sequence_close, device, sizes

from torchrua import C, D, P, segment_head, segment_last, segment_logsumexp, segment_max, segment_mean, segment_min, \
    segment_prod, segment_sum


def reduce_max(tensor: Tensor) -> Tensor:
    return tensor.max(dim=0).values


def reduce_min(tensor: Tensor) -> Tensor:
    return tensor.min(dim=0).values


def reduce_sum(tensor: Tensor) -> Tensor:
    return tensor.sum(dim=0)


def reduce_mean(tensor: Tensor) -> Tensor:
    return tensor.mean(dim=0)


def reduce_prod(tensor: Tensor) -> Tensor:
    return tensor.prod(dim=0)


def reduce_logsumexp(tensor: Tensor) -> Tensor:
    return tensor.logsumexp(dim=0)


def reduce_head(tensor: Tensor) -> Tensor:
    return tensor[0]


def reduce_last(tensor: Tensor) -> Tensor:
    return tensor[-1]


def raw_segment(sequence, duration, fn):
    expected = []

    for index, duration in enumerate(duration):
        start, end, seq = 0, 0, []
        for size in duration:
            start, end = end, end + size
            seq.append(fn(sequence[index][start:end]))

        seq = torch.stack(seq, dim=0)
        expected.append(seq)

    return expected


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
    fns=st.sampled_from([
        (segment_max, reduce_max),
        (segment_min, reduce_min),
        (segment_sum, reduce_sum),
        (segment_mean, reduce_mean),
        (segment_prod, reduce_prod),
        (segment_logsumexp, reduce_logsumexp),
        (segment_head, reduce_head),
        (segment_last, reduce_last),
    ]),
    rua_sequence=st.sampled_from([C.new, D.new, P.new]),
    rua_duration=st.sampled_from([C.new, D.new, P.new]),
)
def test_segment_sequence(token_sizes, dim, fns, rua_sequence, rua_duration):
    inputs = [
        torch.randn((token_size, dim), device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    durations = [
        torch.unique(torch.randint(token_size, (token_size,), device=device), sorted=False, return_counts=True)[1]
        for token_size in token_sizes
    ]

    fn1, fn2 = fns

    actual = rua_sequence(inputs).seg(rua_duration(durations), fn1).cat()
    expected = C.new(raw_segment(D.new(inputs).data, durations, fn2))

    assert_sequence_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual.data, expected=expected.data, inputs=inputs)



# import torch
# from torch import Tensor
#
# from torchrua import major_sizes_to_ptr
#
#
# @torch.no_grad()
# def major_sizes_to_span_ptr(sizes: Tensor):
#     major_ptr, minor_sizes = major_sizes_to_ptr(sizes=sizes)
#     x_ptr, z_ptr = major_sizes_to_ptr(sizes=minor_sizes + 1)
#
#     return major_ptr[z_ptr], minor_sizes[z_ptr], x_ptr, minor_sizes + 1



from typing import List, Union

import torch
from torch.types import Number

from torchrua.ty import C, D, P, T


def split_cdp(sequence: Union[C, D, P]) -> List[T]:
    data, token_sizes = sequence.cat()

    return torch.split(data, token_sizes.cpu().tolist(), dim=0)


C.split = split_cdp
D.split = split_cdp
P.split = split_cdp


def tolist_cdp(sequence: Union[C, D, P]) -> List[List[Number]]:
    return [tensor.tolist() for tensor in sequence.detach().cpu().split()]


C.tolist = tolist_cdp
D.tolist = tolist_cdp
P.tolist = tolist_cdp



import torch
from hypothesis import given, settings, strategies as st
from torchnyan import BATCH_SIZE, FEATURE_DIM, TINY_TOKEN_SIZE, assert_close, assert_grad_close, device, sizes

from torchrua import C, scatter_logsumexp, scatter_max, scatter_mean, scatter_min, scatter_prod, scatter_sum, \
    segment_head, segment_last, segment_logsumexp, segment_max, segment_mean, segment_min, segment_prod, segment_sum


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
    include_self=st.booleans(),
)
def test_scatter_max(token_sizes, dim, include_self):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    index = [
        torch.full((token_size,), fill_value=index, dtype=torch.long, device=device)
        for index, token_size in enumerate(token_sizes)
    ]
    index = torch.cat(index, dim=0)
    permutation = torch.randperm(sum(token_sizes), dtype=torch.long, device=device)

    tensor = torch.randn((len(token_sizes), dim), dtype=torch.float32, device=device)
    expected = torch.cat([torch.max(seq, dim=0, keepdim=True).values for seq in inputs], dim=0)
    if include_self:
        expected = torch.maximum(expected, tensor)

    source, _ = C.new(inputs)
    actual = scatter_max(tensor, index=index[permutation], source=source[permutation], include_self=include_self)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
    include_self=st.booleans(),
)
def test_scatter_min(token_sizes, dim, include_self):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    index = [
        torch.full((token_size,), fill_value=index, dtype=torch.long, device=device)
        for index, token_size in enumerate(token_sizes)
    ]
    index = torch.cat(index, dim=0)
    permutation = torch.randperm(sum(token_sizes), dtype=torch.long, device=device)

    tensor = torch.randn((len(token_sizes), dim), dtype=torch.float32, device=device)
    expected = torch.cat([torch.min(seq, dim=0, keepdim=True).values for seq in inputs], dim=0)
    if include_self:
        expected = torch.minimum(expected, tensor)

    source, _ = C.new(inputs)
    actual = scatter_min(tensor, index=index[permutation], source=source[permutation], include_self=include_self)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
    include_self=st.booleans(),
)
def test_scatter_sum(token_sizes, dim, include_self):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    index = [
        torch.full((token_size,), fill_value=index, dtype=torch.long, device=device)
        for index, token_size in enumerate(token_sizes)
    ]
    index = torch.cat(index, dim=0)
    permutation = torch.randperm(sum(token_sizes), dtype=torch.long, device=device)

    tensor = torch.randn((len(token_sizes), dim), dtype=torch.float32, device=device)
    expected = torch.cat([torch.sum(seq, dim=0, keepdim=True) for seq in inputs], dim=0)
    if include_self:
        expected = expected + tensor

    source, _ = C.new(inputs)
    actual = scatter_sum(tensor, index=index[permutation], source=source[permutation], include_self=include_self)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
    include_self=st.booleans(),
)
def test_scatter_mean(token_sizes, dim, include_self):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    index = [
        torch.full((token_size,), fill_value=index, dtype=torch.long, device=device)
        for index, token_size in enumerate(token_sizes)
    ]
    index = torch.cat(index, dim=0)
    permutation = torch.randperm(sum(token_sizes), dtype=torch.long, device=device)

    token_sizes = torch.tensor(token_sizes, dtype=torch.float32, device=device)

    tensor = torch.randn((len(token_sizes), dim), dtype=torch.float32, device=device)
    expected = torch.cat([torch.sum(seq, dim=0, keepdim=True) for seq in inputs], dim=0)

    if not include_self:
        expected = expected / token_sizes[:, None]
    else:
        expected = (expected + tensor) / (token_sizes[:, None] + 1)

    source, _ = C.new(inputs)
    actual = scatter_mean(tensor, index=index[permutation], source=source[permutation], include_self=include_self)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
    include_self=st.booleans(),
)
def test_scatter_prod(token_sizes, dim, include_self):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    index = [
        torch.full((token_size,), fill_value=index, dtype=torch.long, device=device)
        for index, token_size in enumerate(token_sizes)
    ]
    index = torch.cat(index, dim=0)
    permutation = torch.randperm(sum(token_sizes), dtype=torch.long, device=device)

    tensor = torch.randn((len(token_sizes), dim), dtype=torch.float32, device=device)
    expected = torch.cat([torch.prod(seq, dim=0, keepdim=True) for seq in inputs], dim=0)
    if include_self:
        expected = expected * tensor

    source, _ = C.new(inputs)
    actual = scatter_prod(tensor, index=index[permutation], source=source[permutation], include_self=include_self)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
    include_self=st.booleans(),
)
def test_scatter_logsumexp(token_sizes, dim, include_self):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    index = [
        torch.full((token_size,), fill_value=index, dtype=torch.long, device=device)
        for index, token_size in enumerate(token_sizes)
    ]
    index = torch.cat(index, dim=0)
    permutation = torch.randperm(sum(token_sizes), dtype=torch.long, device=device)

    tensor = torch.randn((len(token_sizes), dim), dtype=torch.float32, device=device)
    expected = torch.cat([torch.logsumexp(seq, dim=0, keepdim=True) for seq in inputs], dim=0)
    if include_self:
        expected = torch.logaddexp(expected, tensor)

    source, _ = C.new(inputs)
    actual = scatter_logsumexp(
        tensor, include_self=include_self,
        index=index[permutation], source=source[permutation],
    )

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_segment_max(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected = torch.cat([torch.max(seq, dim=0, keepdim=True).values for seq in inputs], dim=0)

    tensor, segment_sizes = C.new(inputs)
    actual = segment_max(tensor, segment_sizes=segment_sizes)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_segment_min(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected = torch.cat([torch.min(seq, dim=0, keepdim=True).values for seq in inputs], dim=0)

    tensor, segment_sizes = C.new(inputs)
    actual = segment_min(tensor, segment_sizes=segment_sizes)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_segment_sum(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected = torch.cat([torch.sum(seq, dim=0, keepdim=True) for seq in inputs], dim=0)

    tensor, segment_sizes = C.new(inputs)
    actual = segment_sum(tensor, segment_sizes=segment_sizes)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_segment_mean(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected = torch.cat([torch.mean(seq, dim=0, keepdim=True) for seq in inputs], dim=0)

    tensor, segment_sizes = C.new(inputs)
    actual = segment_mean(tensor, segment_sizes=segment_sizes)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_segment_prod(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected = torch.cat([torch.prod(seq, dim=0, keepdim=True) for seq in inputs], dim=0)

    tensor, segment_sizes = C.new(inputs)
    actual = segment_prod(tensor, segment_sizes=segment_sizes)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_segment_logsumexp(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected = torch.cat([torch.logsumexp(seq, dim=0, keepdim=True) for seq in inputs], dim=0)

    tensor, segment_sizes = C.new(inputs)
    actual = segment_logsumexp(tensor, segment_sizes=segment_sizes)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_segment_head(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected = torch.cat([seq[:1] for seq in inputs], dim=0)

    tensor, segment_sizes = C.new(inputs)
    actual = segment_head(tensor, segment_sizes=segment_sizes)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_segment_last(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), dtype=torch.float32, device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected = torch.cat([seq[-1:] for seq in inputs], dim=0)

    tensor, segment_sizes = C.new(inputs)
    actual = segment_last(tensor, segment_sizes=segment_sizes)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)



import torch
from hypothesis import given, settings, strategies as st
from torchnyan import BATCH_SIZE, FEATURE_DIM, TOKEN_SIZE, assert_grad_close, assert_sequence_close, device, sizes

from torchrua import C, D, P


@settings(deadline=None)
@given(
    data=st.data(),
    token_sizes=sizes(BATCH_SIZE, TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
    rua_sequence=st.sampled_from([C.new, D.new, P.new]),
)
def test_roll_sequence(data, token_sizes, dim, rua_sequence):
    shifts = data.draw(st.integers(min_value=-max(token_sizes), max_value=+max(token_sizes)))

    inputs = [
        torch.randn((token_size, dim), device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    actual = rua_sequence(inputs).roll(shifts=shifts).cat()
    expected = C.new([sequence.roll(shifts, dims=[0]) for sequence in inputs])

    assert_sequence_close(actual, expected)
    assert_grad_close(actual.data, expected.data, inputs=inputs)



import torch
from hypothesis import given, settings
from torch.nn.utils.rnn import pack_sequence, pad_sequence
from torchnyan import BATCH_SIZE, FEATURE_DIM, TOKEN_SIZE, assert_grad_close, assert_sequence_close, device, sizes

from torchrua import C, D


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_cat_packed_sequence(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    actual = C.new(inputs)
    expected = pack_sequence(inputs, enforce_sorted=False).cat()

    assert_sequence_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual.data, expected=expected.data, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
)
def test_cat_padded_sequence(token_sizes, dim):
    inputs = [
        torch.randn((token_size, dim), device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    actual = C.new(inputs)
    expected = D(
        data=pad_sequence(inputs, batch_first=True),
        token_sizes=torch.tensor(token_sizes, device=device),
    ).cat()

    assert_sequence_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual.data, expected=expected.data, inputs=inputs)



from typing import Tuple

import torch
from torch import Tensor


def major_sizes_to_ptr(sizes: Tensor) -> Tuple[Tensor, Tensor]:
    minor_ptr = torch.repeat_interleave(repeats=sizes)

    major_ptr = torch.repeat_interleave(get_offsets(sizes=sizes), repeats=sizes)
    major_ptr = torch.arange(major_ptr.size()[0], device=major_ptr.device) - major_ptr

    return major_ptr, minor_ptr


def get_offsets(sizes: Tensor) -> Tensor:
    sizes = sizes.cumsum(dim=0).roll(shifts=1, dims=0)
    sizes[0] = 0
    return sizes


def invert_permutation(tensor: Tensor) -> Tensor:
    index = torch.arange(tensor.size()[0], device=tensor.device)
    permutation = torch.empty_like(index)
    permutation[tensor] = index
    return permutation


def _self(self, *_, **__):
    return self



from typing import List

import torch
from torch import Tensor


def with_shape(shape: torch.Size, dim: int, value: int) -> List[int]:
    shape = list(shape)
    shape[dim] = value
    return shape


def broadcast_shapes(*sizes: torch.Size, dim: int):
    shape = torch.broadcast_shapes(*[with_shape(size, dim=dim, value=1) for size in sizes])
    return [with_shape(shape, dim=dim, value=size[dim]) for size in sizes]


def broadcast_tensors(*tensors: Tensor, dim: int):
    shapes = broadcast_shapes(*[tensor.size() for tensor in tensors], dim=dim)
    return [tensor.expand(shape) for tensor, shape in zip(tensors, shapes)]


def gather(tensor: Tensor, index: Tensor, dim: int) -> Tensor:
    tensor, index = broadcast_tensors(tensor, index, dim=dim)
    return tensor.gather(dim=dim, index=index)



from typing import List

import torch

from torchrua.core import _self
from torchrua.ty import C, D, P, T


def cat_sequence(sequence: List[T]) -> C:
    data = torch.cat(sequence, dim=0)
    token_sizes = [s.size()[0] for s in sequence]
    return C(data=data, token_sizes=data.new_tensor(token_sizes, dtype=torch.long))


C.new = cat_sequence
C.cat = _self


def cat_t(sequence: T) -> C:
    token_sizes = sequence.new_tensor(sequence.size()[:1], dtype=torch.long)
    return C(data=sequence, token_sizes=token_sizes)


T.cat = cat_t


def cat_d(sequence: D) -> C:
    return sequence.idx().rua(sequence)


D.cat = cat_d


def cat_p(sequence: P) -> C:
    data, batch_sizes, sorted_indices, unsorted_indices = sequence
    b, t, *sizes = sequence.size()

    if len(sizes) > 0:
        return sequence.idx().cat().rua(sequence)

    batch_ptr, token_ptr = sequence.ptr()
    batch_ptr = sorted_indices[batch_ptr]

    tensor = data.new_zeros((b, t))
    tensor[batch_ptr, token_ptr] = data

    mask = torch.zeros_like(tensor, dtype=torch.long)
    mask[batch_ptr, token_ptr] = 1

    return C(
        data=tensor[mask.bool()],
        token_sizes=mask.sum(dim=1),
    )


P.cat = cat_p



import torch
from hypothesis import given, settings, strategies as st
from torch.nn.utils.rnn import pad_sequence
from torchnyan import BATCH_SIZE, TOKEN_SIZE, assert_close, device, sizes

from torchrua import C, D, P


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TOKEN_SIZE),
    rua_sequence=st.sampled_from([C.new, D.new, P.new]),
    zero_one_dtype=st.sampled_from([
        (False, True, torch.bool),
        (-1, +2, torch.long),
        (torch.finfo(torch.float16).min, torch.finfo(torch.float16).max, torch.float16),
        (torch.finfo(torch.float32).min, torch.finfo(torch.float32).max, torch.float32),
        (torch.finfo(torch.float64).min, torch.finfo(torch.float64).max, torch.float64),
    ])
)
def test_mask_sequence(token_sizes, rua_sequence, zero_one_dtype):
    inputs = [
        torch.randn((token_size,), device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    zero, one, dtype = zero_one_dtype

    actual = rua_sequence(inputs).mask(zero=zero, one=one, dtype=dtype)
    expected = pad_sequence([
        torch.full((token_size,), fill_value=one, device=device, dtype=dtype)
        for token_size in token_sizes
    ], batch_first=True, padding_value=zero)

    assert_close(actual=actual, expected=expected)



from pathlib import Path

from setuptools import find_packages, setup

name = 'torchrua'

root_dir = Path(__file__).parent.resolve()
with (root_dir / 'requirements.txt').open(mode='r', encoding='utf-8') as fp:
    install_requires = [install_require.strip() for install_require in fp]

setup(
    name=name,
    version='0.5.0',
    packages=[package for package in find_packages() if package.startswith(name)],
    url='https://github.com/speedcell4/torchrua',
    license='MIT',
    author='speedcell4',
    author_email='speedcell4@gmail.com',
    description='rua your variable-length tensors!',
    python_requires='>=3.8',
    install_requires=install_requires,
)



import torch

from torchrua.ty import C, D, P


def rev_c(sequence: C) -> C:
    data, token_sizes = sequence
    _, _, *sizes = sequence.size()

    if len(sizes) > 0:
        return sequence.idx().rev().rua(sequence)

    data = torch.flip(data, dims=[0])
    token_sizes = torch.flip(token_sizes, dims=[0])

    data = torch.split(data, token_sizes.detach().cpu().tolist(), dim=0)
    return sequence._replace(data=torch.cat(data[::-1], dim=0))


C.rev = rev_c


def rev_d(sequence: D) -> D:
    index1, _ = idx = sequence.idx()
    index2, _ = idx.rev()

    data = sequence._data().clone()
    data[index1] = data[index2]

    return sequence._replace(data=data.view_as(sequence.data))


D.rev = rev_d


def rev_p(sequence: P) -> P:
    return sequence.idx().cat().rev().pack().rua(sequence)


P.rev = rev_p



from typing import Union

import torch

from torchrua.ty import C, D, P


def seg_c(sequence: C, duration: Union[C, D, P], fn) -> C:
    data, token_sizes = sequence
    duration = duration.cat()

    data = fn(data, duration.data)
    return duration._replace(data=data)


C.seg = seg_c


def seg_d(sequence: D, duration: Union[C, D, P], fn) -> D:
    duration, token_sizes = duration.pad(fill_value=0)

    remaining = sequence.size()[1] - duration.sum(dim=1, keepdim=True)
    duration = torch.cat([duration, remaining], dim=-1)

    data = fn(sequence._data(), duration.view(-1))
    data = data.view((*duration.size(), *data.size()[1:]))
    return D(data=data[:, :-1], token_sizes=token_sizes)


D.seg = seg_d


def seg_p(sequence: P, duration: Union[C, D, P], fn) -> P:
    return sequence.cat().seg(duration, fn).pack()


P.seg = seg_p



import torch
from hypothesis import given, settings, strategies as st
from torchnyan import BATCH_SIZE, FEATURE_DIM, TOKEN_SIZE, assert_close, assert_grad_close, device, sizes

from torchrua import C, D, P


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TOKEN_SIZE),
    dim=sizes(FEATURE_DIM),
    rua_sequence=st.sampled_from([C.new, D.new, P.new]),
)
def test_last_sequence(token_sizes, dim, rua_sequence):
    inputs = [
        torch.randn((token_size, dim), device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    actual = rua_sequence(inputs).last()
    expected = torch.stack([sequence[-1] for sequence in inputs], dim=0)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)



from typing import List, Union

import torch

from torchrua import C, D, P
from torchrua.core import invert_permutation


def compose(sequences: List[Union[C, D, P]]) -> P:
    offset, data, indices, token_sizes = 0, [], [], []

    for sequence in sequences:
        data.append(sequence._data())

        idx, sizes = sequence.idx().cat()
        indices.append(idx + offset)
        token_sizes.append(sizes)

        offset += data[-1].size()[0]

    token_sizes = C.new(token_sizes)
    unsorted_indices, _, _, _ = token_sizes.idx().pack()

    indices = C(data=torch.cat(indices, dim=0), token_sizes=token_sizes.data).pack()
    unsorted_indices = indices.unsorted_indices[unsorted_indices]
    sorted_indices = invert_permutation(unsorted_indices)

    indices = indices._replace(
        sorted_indices=sorted_indices,
        unsorted_indices=unsorted_indices,
    )

    return indices.rua(torch.cat(data, dim=0))
